# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for Matter endpoint capability wrapper."""
import abc
from typing import Any, Collection, List, Mapping, Optional, Set, Tuple, Type
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base
import immutabledict

ROOT_NODE_ENDPOINT_ID = 0

logger = gdm_logger.get_logger()


class MatterEndpointsBase(capability_base.CapabilityBase):
  """Capability wrapper for accessing the Matter endpoint instances."""

  def __init__(self, device_name: str, **cluster_kwargs: Any):
    """Initializes an instance of MatterEndpoints capability.

    Args:
      device_name: Name of the device instance the capability is attached to.
      **cluster_kwargs: Keyword arguments for initializing PigweedRPC/ChipTool
        based cluster capability.
    """
    super().__init__(device_name=device_name)
    self._cluster_kwargs = cluster_kwargs

    # Endpoint ID to endpoint instance mapping
    self._endpoints = {}

    # The endpoint ID to endpoint class mapping.
    self._endpoint_id_to_class = {}

    # The endpoint class to endpoint ID mapping.
    self._endpoint_class_to_id = {}

    # The endpoint ID to set of cluster classes mapping.
    self._endpoint_id_to_clusters = {}

    # The endpoint ID to device type ID mapping.
    self._endpoint_id_to_device_type_id = {}

  @abc.abstractmethod
  def get_supported_endpoint_ids(self) -> List[int]:
    """Gets the list of supported endpoint ids on the device."""

  @abc.abstractmethod
  def get_endpoint_class_and_device_type_id(
      self, endpoint_id: int) -> Tuple[Type[endpoint_base.EndpointBase], int]:
    """Gets the endpoint class and device type ID by the given endpoint id.

    Args:
      endpoint_id: The given endpoint ID on the device.

    Returns:
      The endpoint class module (or UnsupportedEndpoint if the endpoint is not
      yet supported in GDM) and the device type ID.
    """

  @abc.abstractmethod
  def get_supported_clusters(
      self, endpoint_id: int) -> Set[Type[cluster_base.ClusterBase]]:
    """Retrieves the supported clusters from the given endpoint ID.

    Args:
      endpoint_id: The given endpoint ID on the device.

    Returns:
      Set of supported cluster capability classes.
    """

  @decorators.DynamicProperty
  def endpoint_id_to_class(
      self) -> Mapping[int, Optional[Type[endpoint_base.EndpointBase]]]:
    """Returns the endpoint_id_to_class mapping."""
    self._fetch_endpoints_and_clusters()
    return immutabledict.immutabledict(self._endpoint_id_to_class)

  @decorators.DynamicProperty
  def endpoint_class_to_id(
      self) -> Mapping[Type[endpoint_base.EndpointBase], int]:
    """Returns the endpoint_class_to_id mapping."""
    self._fetch_endpoints_and_clusters()
    return immutabledict.immutabledict(self._endpoint_class_to_id)

  @decorators.DynamicProperty
  def endpoint_id_to_clusters(
      self) -> Mapping[int, Set[Type[cluster_base.ClusterBase]]]:
    """Returns the endpoint ID to cluster classes mapping."""
    self._fetch_endpoints_and_clusters()
    return immutabledict.immutabledict(self._endpoint_id_to_clusters)

  @decorators.DynamicProperty
  def endpoint_id_to_device_type_id(self) -> Mapping[int, int]:
    """Returns the endpoint ID to device type ID mapping."""
    self._fetch_endpoints_and_clusters()
    return immutabledict.immutabledict(self._endpoint_id_to_device_type_id)

  def _fetch_endpoints_and_clusters(self):
    """Retrieves the supported endpoints and clusters from descriptor cluster.

    The descriptor cluster should only be queried if it has not previously been
    called or the reset method is called.
    """
    if not self._endpoint_id_to_class:
      for endpoint_id in self.get_supported_endpoint_ids():
        endpoint_cls, device_type_id = (
            self.get_endpoint_class_and_device_type_id(endpoint_id))

        # Store the endpoint ID to endpoint class mapping.
        self._endpoint_id_to_class[endpoint_id] = endpoint_cls

        # Store the endpoint ID to device type ID mapping.
        self._endpoint_id_to_device_type_id[endpoint_id] = device_type_id

        # Ensuring we store the first endpoint ID handled by this class.
        # This mapping will be used in get_endpoint_instance_by_class method
        if endpoint_cls not in self._endpoint_class_to_id:
          self._endpoint_class_to_id[endpoint_cls] = endpoint_id

        # Store the endpoint ID to clusters mapping.
        self._endpoint_id_to_clusters[endpoint_id] = (
            self.get_supported_clusters(endpoint_id))

  @decorators.CapabilityLogDecorator(logger)
  def get(self, endpoint_id: int) -> endpoint_base.EndpointBase:
    """Gets the specific endpoint instance by endpoint ID.

    Args:
      endpoint_id: Endpoint ID on the device.

    Returns:
      The endpoint class for the given endpoint ID.

    Raises:
      DeviceError: The given endpoint ID does not exist on the device. Or the
      endpoint class for the given endpoint ID is not implemented yet.
    """
    if endpoint_id not in self.endpoint_id_to_class:
      raise errors.DeviceError(
          f"Endpoint ID {endpoint_id} on {self._device_name} does not exist.")

    if endpoint_id not in self._endpoints:
      endpoint_class = self.endpoint_id_to_class[endpoint_id]
      supported_clusters = self.endpoint_id_to_clusters[endpoint_id]
      device_type_id = self.endpoint_id_to_device_type_id[endpoint_id]

      self._endpoints[endpoint_id] = endpoint_class(
          device_name=self._device_name,
          identifier=endpoint_id,
          device_type_id=device_type_id,
          supported_clusters=frozenset(supported_clusters),
          **self._cluster_kwargs)

    return self._endpoints[endpoint_id]

  @decorators.CapabilityLogDecorator(logger)
  def list(self) -> Mapping[int, Type[endpoint_base.EndpointBase]]:
    """Returns a mapping of endpoint ID to the supported endpoint class."""
    return self.endpoint_id_to_class

  @decorators.CapabilityLogDecorator(logger)
  def get_endpoint_instance_by_class(
      self, endpoint_class: Type[endpoint_base.EndpointBase]
  ) -> endpoint_base.EndpointBase:
    """Gets the endpoint instance by the given endpoint class.

    Args:
      endpoint_class: The given Matter endpoint class.

    Raises:
      DeviceError: When the given endpoint class is not supported on the device.

    Returns:
      The endpoint instance.
    """
    if endpoint_class not in self.endpoint_class_to_id:
      raise errors.DeviceError(
          f"Class {endpoint_class} is not supported on {self._device_name}.")
    endpoint_id = self.endpoint_class_to_id[endpoint_class]
    return self.get(endpoint_id)

  @decorators.CapabilityLogDecorator(logger)
  def reset(self) -> None:
    """Resets the endpoint ID and endpoint class mappings."""
    self._endpoint_id_to_class.clear()
    self._endpoint_class_to_id.clear()
    self._endpoint_id_to_clusters.clear()
    self._endpoint_id_to_device_type_id.clear()
    self._endpoints.clear()

  @decorators.CapabilityLogDecorator(logger)
  def has_endpoints(self, endpoint_names: Collection[str]) -> bool:
    """Checks whether the device supports all the given endpoint names.

    Args:
      endpoint_names: The collection of endpoint names. The names are
        case-insensitive. Some valid examples are: "on_off_light",
        "On_Off_Light".

    Returns:
      True if the device supports all the endpoints, false otherwise.
    """
    valid_endpoint_name_to_class = {
        endpoint_class.get_capability_name(): endpoint_class
        for endpoint_class in self._SUPPORTED_ENDPOINTS
    }
    supported_endpoints = set(self.list().values())
    for endpoint_name in endpoint_names:
      endpoint = valid_endpoint_name_to_class.get(endpoint_name.lower())
      if endpoint is None:
        logger.warning(f"Endpoint {endpoint_name} is not recognized. "
                       "Valid endpoints are: "
                       f"{list(valid_endpoint_name_to_class.keys())}")
        return False
      if endpoint not in supported_endpoints:
        return False
    return True

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoints(self) -> List[str]:
    """Returns names of endpoints supported by the device."""
    return sorted(endpoint.get_capability_name()
                  for endpoint in self.get_supported_endpoint_flavors())

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoint_flavors(
      self) -> List[Type[endpoint_base.EndpointBase]]:
    """Returns flavors of endpoints supported by the device."""
    return [
        endpoint for endpoint in self.list().values() if endpoint is not None
    ]

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoints_and_clusters(self) -> Mapping[int, Set[str]]:
    """Returns the supported endpoint IDs and set of cluster names mapping."""
    return {
        endpoint_id: self.get(endpoint_id).get_supported_clusters()
        for endpoint_id in self.list()
    }

  @decorators.CapabilityLogDecorator(logger)
  def get_supported_endpoint_instances_and_cluster_flavors(
      self
  ) -> Mapping[Type[endpoint_base.EndpointBase],
               Set[Type[cluster_base.ClusterBase]]]:
    """Returns the supported endpoint instance and cluster flavors mapping."""
    mapping = {}
    for endpoint_id in self.list():
      endpoint = self.get(endpoint_id)
      mapping[endpoint] = endpoint.get_supported_cluster_flavors()
    return mapping



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""chip-tool implementation of the Matter Door Lock cluster capability.
"""
from typing import Literal

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import door_lock_base

logger = gdm_logger.get_logger()

_CLUSTER_NAME = "doorlock"


class DoorLockClusterChipTool(door_lock_base.DoorLockClusterBase):
  """chip-tool based Matter Door Lock cluster capability."""

  @decorators.CapabilityLogDecorator(logger)
  def lock_door(self, verify: bool = True) -> None:
    """The Lock Door command to lock the Matter device.

    Args:
      verify: If true, verifies the state configurations before returning.
    """
    self._lock_command("lock-door", verify)

  @decorators.CapabilityLogDecorator(logger)
  def unlock_door(self, verify: bool = True) -> None:
    """The Unlock Door command to unlock the Matter device.

    Args:
      verify: If true, verifies the state configurations before returning.
    """
    self._lock_command("unlock-door", verify)

  @decorators.DynamicProperty
  def lock_state(self) -> matter_enums.LockState:
    """The LockState attribute."""
    locked_data = self._read(self._endpoint_id, _CLUSTER_NAME, "lock-state")
    return matter_enums.LockState(int(locked_data))

  def _lock_command(
      self,
      lock_command: Literal["lock-door", "unlock-door"],
      verify: bool = True) -> None:
    """Locks or unlocks the device.

    Args:
      lock_command: "lock-door" or "unlock-door".
      verify: If true, verifies the lock configurations before returning.

    Raises:
      DeviceError: when the device does not transition to the appropriate
      lock state.
    """
    self._send(
        endpoint_id=self._endpoint_id,
        cluster=_CLUSTER_NAME,
        command=lock_command,
        arguments=[],
        flags=["--timedInteractionTimeoutMs", 1000])

    if verify:
      expected_lock_state = (
          matter_enums.LockState.LOCKED if lock_command == "lock-door" else
          matter_enums.LockState.UNLOCKED)
      if expected_lock_state != self.lock_state:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            f"Device {self._device_name} lock state attribute did not change "
            f"from {self.lock_state}.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RPC implementation of Matter Window Covering cluster capability.
"""

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import window_covering_base
from gazoo_device.protos import attributes_service_pb2


logger = gdm_logger.get_logger()
_WINDOW_COVERING_CLUSTER = matter_enums.WindowCoveringCluster
# TODO(b/241698162): switch to percent when TLV is supported.
PERCENT_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_INT8U_ATTRIBUTE_TYPE


class WindowCoveringClusterPwRpc(window_covering_base.WindowCoveringClusterBase
                                ):
  """Matter Matter Window Covering cluster capability."""

  @decorators.DynamicProperty
  def current_position_lift_percentage(self) -> int:
    """The CurrentPositionLiftPercentage attribute.

    The CurrentPositionLiftPercentage attribute identifies the actual position
    as a percentage from 0% to 100% with 1% default step. This attribute is
    equal to CurrentPositionLiftPercent100ths attribute divided by 100.

    Returns:
      The CurrentPositionLiftPercentage attribute.
    """
    return self._read_percent_attribute(
        attribute_id=_WINDOW_COVERING_CLUSTER
        .ATTRIBUTE_CURRENT_POSITION_LIFT_PERCENTAGE)

  @current_position_lift_percentage.setter
  def current_position_lift_percentage(self, percent: int) -> None:
    """Updates the CurrentPositionLiftPercentage attribute.

    The CurrentPositionLiftPercentage attribute identifies the actual position
    as a percentage from 0% to 100% with 1% default step. This attribute is
    equal to CurrentPositionLiftPercent100ths attribute divided by 100.

    Args:
      percent: The value CurrentPositionLiftPercentage should update to.
    """
    self._write_percent_attribute(attribute_id=_WINDOW_COVERING_CLUSTER
                                  .ATTRIBUTE_CURRENT_POSITION_LIFT_PERCENTAGE,
                                  percent=percent)

    if self.current_position_lift_percentage != percent:  # pylint: disable=comparison-with-callable
      raise errors.DeviceError(
          f"Device {self._device_name} current_position_lift_percentage didn't "
          f"change to {percent}.")

  @decorators.DynamicProperty
  def current_position_tilt_percentage(self) -> int:
    """The CurrentPositionTiltPercentage attribute.

    The CurrentPositionTiltPercentage attribute identifies the actual position
    as a percentage from 0% to 100% with 1% default step. This attribute is
    equal to CurrentPositionTiltPercent100ths attribute divided by 100.

    Returns:
      The CurrentPositionTiltPercentage attribute.
    """
    return self._read_percent_attribute(
        attribute_id=_WINDOW_COVERING_CLUSTER
        .ATTRIBUTE_CURRENT_POSITION_TILT_PERCENTAGE)

  @current_position_tilt_percentage.setter
  def current_position_tilt_percentage(self, percent: int) -> None:
    """Updates the CurrentPositionTiltPercentage attribute.

    The CurrentPositionTiltPercentage attribute identifies the actual position
    as a percentage from 0% to 100% with 1% default step. This attribute is
    equal to CurrentPositionTiltPercent100ths attribute divided by 100.

    Args:
      percent: The value CurrentPositionTiltPercentage should update to.
    """
    self._write_percent_attribute(attribute_id=_WINDOW_COVERING_CLUSTER
                                  .ATTRIBUTE_CURRENT_POSITION_TILT_PERCENTAGE,
                                  percent=percent)

    if self.current_position_tilt_percentage != percent:  # pylint: disable=comparison-with-callable
      raise errors.DeviceError(
          f"Device {self._device_name} current_position_tilt_percentage didn't "
          f"change to {percent}.")

  def _read_percent_attribute(self, attribute_id: int) -> int:
    """Helper method to read percent attribute type."""
    measured_value_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_WINDOW_COVERING_CLUSTER.ID,
        attribute_id=attribute_id,
        attribute_type=PERCENT_ATTRIBUTE_TYPE)
    return measured_value_data.data_uint8

  def _write_percent_attribute(self, attribute_id: int, percent: int) -> None:
    """Helper method to write percent attribute type."""
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=_WINDOW_COVERING_CLUSTER.ID,
        attribute_id=attribute_id,
        attribute_type=PERCENT_ATTRIBUTE_TYPE,
        data_uint8=percent)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter spec enum module."""
import enum
from gazoo_device.protos import attributes_service_pb2

# The Matter cluster enums definitions: (only enums used in GDM are defined)


class ColorControlCluster(enum.IntEnum):
  """Color control cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_COLOR_CONTROL_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_CURRENT_HUE = 0x0000
  ATTRIBUTE_CURRENT_SATURATION = 0x0001
  ATTRIBUTE_COLOR_TEMPERATURE_MIREDS = 0x0007


class DoorLockCluster(enum.IntEnum):
  """Door lock cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_DOOR_LOCK_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_LOCK_STATE = 0x0000


class FlowMeasurementCluster(enum.IntEnum):
  """Flow Measurement cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_FLOW_MEASUREMENT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_MEASURED_VALUE = 0x0000
  ATTRIBUTE_MIN_MEASURED_VALUE = 0x0001
  ATTRIBUTE_MAX_MEASURED_VALUE = 0x0002


class FanControlCluster(enum.IntEnum):
  """Fan Control cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_FAN_CONTROL_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_FAN_MODE = 0x0000
  ATTRIBUTE_FAN_MODE_SEQUENCE = 0x0001
  ATTRIBUTE_PERCENT_SETTING = 0x0002
  ATTRIBUTE_PERCENT_CURRENT = 0x0003
  ATTRIBUTE_SPEED_MAX = 0x0004
  ATTRIBUTE_SPEED_SETTING = 0x0005
  ATTRIBUTE_SPEED_CURRENT = 0x0006


class FanMode(enum.IntEnum):
  """Fan Mode attribute enum.
  """
  OFF = 0
  LOW = 1
  MEDIUM = 2
  HIGH = 3
  ON = 4
  AUTO = 5
  SMART = 6


class FanModeSequence(enum.IntEnum):
  """Fan Mode Sequence attribute enum.
  """
  OFF_LOW_MED_HIGH = 0
  OFF_LOW_HIGH = 1
  OFF_LOW_MED_HIGH_AUTO = 2
  OFF_LOW_HIGH_AUTO = 3
  OFF_ON_AUTO = 4
  OFF_ON = 5


class IlluminanceMeasurementCluster(enum.IntEnum):
  """Illuminance Measurement cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_ILLUMINANCE_MEASUREMENT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_MEASURED_VALUE = 0x0000
  ATTRIBUTE_MIN_MEASURED_VALUE = 0x0001
  ATTRIBUTE_MAX_MEASURED_VALUE = 0x0002
  ATTRIBUTE_LIGHT_SENSOR_TYPE = 0x0004


class LightSensorType(enum.IntEnum):
  """LightSensorType attribute.
  """
  PHOTODIODE = 0
  CMOS = 1


class LevelControlCluster(enum.IntEnum):
  """Level control cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_LEVEL_CONTROL_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_CURRENT_LEVEL = 0x0000
  ATTRIBUTE_MIN_LEVEL = 0x0002
  ATTRIBUTE_MAX_LEVEL = 0x0003


class OccupancySensingCluster(enum.IntEnum):
  """Occupacny sensing cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_OCCUPANCY_SENSING_CLUSTER_ID

  # Attributes ID
  ATTRIBUTE_OCCUPANCY = 0x0000
  ATTRIBUTE_OCCUPANCY_SENSOR_TYPE = 0x0001
  ATTRIBUTE_OCCUPANCY_SENSOR_TYPE_BITMAP = 0x0002


class OnOffCluster(enum.IntEnum):
  """OnOff cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_ON_OFF_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_ON_OFF = 0x0000


class PressureMeasurementCluster(enum.IntEnum):
  """Pressure Measurement cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_PRESSURE_MEASUREMENT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_MEASURED_VALUE = 0
  ATTRIBUTE_MIN_MEASURED_VALUE = 1
  ATTRIBUTE_MAX_MEASURED_VALUE = 2


class RelativeHumidityMeasurementCluster(enum.IntEnum):
  """Relative Humidity Measurement cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_RELATIVE_HUMIDITY_MEASUREMENT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_MEASURED_VALUE = 0
  ATTRIBUTE_MIN_MEASURED_VALUE = 1
  ATTRIBUTE_MAX_MEASURED_VALUE = 2


class TemperatureMeasurementCluster(enum.IntEnum):
  """Temperature Measurement cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_TEMP_MEASUREMENT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_MEASURED_VALUE = 0
  ATTRIBUTE_MIN_MEASURED_VALUE = 1
  ATTRIBUTE_MAX_MEASURED_VALUE = 2


# Lock state attribute enums
class LockState(enum.IntEnum):
  """Lock state attribute values.
  """
  NOT_FULLY_LOCKED = 0
  LOCKED = 1
  UNLOCKED = 2


# Occupancy Sensor type enums.
class OccupancySensorType(enum.IntEnum):
  """Occupancy Sensor type.
  """
  PIR = 0
  ULTRASONIC = 1
  PIR_AND_ULTRASONIC = 2
  PHYSICAL_CONTACT = 3


class BooleanStateCluster(enum.IntEnum):
  """Boolean State cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_BOOLEAN_STATE_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_STATE_VALUE = 0


class ThermostatCluster(enum.IntEnum):
  """Thermostat cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_THERMOSTAT_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_LOCAL_TEMPERATURE = 0x0000
  ATTRIBUTE_OCCUPIED_COOLING_SETPOINT = 0x0011
  ATTRIBUTE_OCCUPIED_HEATING_SETPOINT = 0x0012
  ATTRIBUTE_CONTROL_SEQUENCE_OF_OPERATION = 0x001b
  ATTRIBUTE_SYSTEM_MODE = 0x001c


class ThermostatControlSequence(enum.IntEnum):
  """Thermostat Control Sequence.

  The enum values are defined in the Matter spec.
  """

  COOLING_ONLY = 0
  COLLING_WITH_REHEAT = 1
  HEATING_ONLY = 2
  HEATING_WITH_REHEAT = 3
  COOLING_AND_HEATING = 4
  COOLING_AND_HEATING_WITH_REHEAT = 5


class ThermostatSystemMode(enum.IntEnum):
  """Thermostat System Mode.

  The enum values are defined in the Matter spec.
  """

  OFF = 0
  AUTO = 1
  COOL = 3
  HEAT = 4
  EMERGENCY_HEAT = 5
  PRECOOLING = 6
  FAN_ONLY = 7
  DRY = 8
  SLEEP = 9


class ThermostatSetpointMode(enum.IntEnum):
  """Thermostat Setpoint Mode.

  The enum values are defined in the Matter spec.
  """

  HEAT = 0
  COOL = 1
  BOTH = 2


class BasicInformationCluster(enum.IntEnum):
  """Basic information cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_BASIC_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_DATA_MODEL_REVISION = 0x0000
  ATTRIBUTE_VENDOR_NAME = 0x0001
  ATTRIBUTE_VENDOR_ID = 0x0002
  ATTRIBUTE_PRODUCT_NAME = 0x0003
  ATTRIBUTE_PRODUCT_ID = 0x0004
  ATTRIBUTE_NODE_LABEL = 0x0005
  ATTRIBUTE_LOCATION = 0x0006
  ATTRIBUTE_HARDWARE_VERSION = 0x0007
  ATTRIBUTE_HARDWARE_VERSION_STRING = 0x008
  ATTRIBUTE_SOFTWARE_VERSION = 0x009
  ATTRIBUTE_SOFTWARE_VERSION_STRING = 0x00A


class WindowCoveringCluster(enum.IntEnum):
  """Window Covering cluster ID and its attribute IDs.

  The enum values are defined in the Matter spec.
  """
  ID = attributes_service_pb2.ClusterType.ZCL_WINDOW_COVERING_CLUSTER_ID

  # Attribute IDs
  ATTRIBUTE_CURRENT_POSITION_LIFT_PERCENTAGE = 0x0008
  ATTRIBUTE_CURRENT_POSITION_TILT_PERCENTAGE = 0x0009




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the Pigweed RPC device common capability."""
import time
from typing import Any, Callable, Dict, List, Optional, Tuple
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_common_base
from gazoo_device.protos import device_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport
from gazoo_device.utility import retry


logger = gdm_logger.get_logger()
_POLL_INTERVAL_SEC = 0.5  # seconds
_WAIT_FOR_DEVICE_READY_SEC = 1  # second
_DEFAULT_BOOTUP_TIMEOUT = 30  # seconds
# TODO(b/235632665) Remove this once the sample app has a default fw version.
_DEFAULT_SW_VERSION = "0"


class PwRPCCommonDefault(pwrpc_common_base.PwRPCCommonBase):
  """Pigweed RPC common capability for devices communicating over PwRPC."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any],
               rpc_timeout_s: int,
               pigweed_port: int = 0):
    """Create an instance of the PwRPCCommonDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method which calls to the endpoint.
      rpc_timeout_s: Timeout (s) for RPC call.
      pigweed_port: Pigweed RPC transport port number.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call
    self._rpc_timeout_s = rpc_timeout_s
    self._pigweed_port = pigweed_port

  @decorators.DynamicProperty
  def software_version(self) -> str:
    """The software version of the device."""
    version = self.get_device_info().software_version_string
    return version or _DEFAULT_SW_VERSION

  @decorators.DynamicProperty
  def qr_code(self) -> str:
    """The pairing QR code of the device."""
    return self.pairing_info.qr_code

  @decorators.DynamicProperty
  def qr_code_url(self) -> str:
    """The pairing QR code URL of the device."""
    return self.pairing_info.qr_code_url

  @decorators.DynamicProperty
  def pairing_state(self) -> bool:
    """The pairing state of the device.

    The device is paired if it has at least one fabric ID and the fabric
    contains a node ID.

    Returns:
      The pairing state.
    """
    return bool(self.fabric_info) and hasattr(self.fabric_info[0], "node_id")

  @decorators.DynamicProperty
  def pairing_info(self) -> device_service_pb2.PairingInfo:
    """The pairing information of the device."""
    return self.get_device_info().pairing_info

  @decorators.DynamicProperty
  def fabric_info(self) -> List[device_service_pb2.FabricInfo]:
    """The list of fabric information of the device."""
    repeated_fabric_info = self.get_device_state().fabric_info
    return list(repeated_fabric_info)

  @decorators.CapabilityLogDecorator(logger)
  def call(self,
           method_name: str,
           method_args: Tuple[Any, ...] = (),
           method_kwargs: Optional[Dict[str, Any]] = None) -> Any:
    """Switchboard RPC call wrapper with specific port number."""
    return self._switchboard_call(
        method_name=method_name,
        method_args=method_args,
        method_kwargs=method_kwargs,
        port=self._pigweed_port)

  @decorators.CapabilityLogDecorator(logger)
  def reboot(self,
             verify: bool = True,
             bootup_timeout_s: int = _DEFAULT_BOOTUP_TIMEOUT) -> None:
    """Reboots the device.

    Args:
      verify: If true, waits for device bootup completes before returning.
      bootup_timeout_s: Timeout (s) to wait for the device to boot up.
    """
    self._trigger_device_action(action="Reboot")
    if verify:
      self.wait_for_bootup_complete(bootup_timeout_s)

  @decorators.CapabilityLogDecorator(logger)
  def factory_reset(self,
                    verify: bool = True,
                    bootup_timeout_s: int = _DEFAULT_BOOTUP_TIMEOUT) -> None:
    """Factory resets the device.

    Args:
      verify: If true, waits for device bootup completes before returning.
      bootup_timeout_s: Timeout (s) to wait for the device to boot up.
    """
    self._trigger_device_action(action="FactoryReset")
    if verify:
      self.wait_for_bootup_complete(bootup_timeout_s)

  @decorators.CapabilityLogDecorator(logger)
  def ota(self) -> None:
    """Triggers OTA to the device."""
    self._trigger_device_action(action="TriggerOta")

  @decorators.CapabilityLogDecorator(logger)
  def wait_for_bootup_complete(self, bootup_timeout: int) -> None:
    """Waits for device to boot up.

    Args:
      bootup_timeout: Max time to wait for bootup to complete (in seconds).

    Raises:
      DeviceError: If device did not boot up successfully in given timeout.
    """
    # For b/272311935, devices will enter bad state (unresponsive to RPC) when
    # trying to send the RPC requests right after rebooting or FDR.
    # Adding 1 second to wait for the device to be in a ready state before
    # sending RPC for polling.
    time.sleep(_WAIT_FOR_DEVICE_READY_SEC)
    try:
      start_time = time.time()
      retry.retry(
          func=self.get_device_info,
          timeout=bootup_timeout,
          interval=_POLL_INTERVAL_SEC,
          reraise=False)
      logger.info(
          f"{self._device_name} booted up in {time.time() - start_time}s.")
    except errors.CommunicationTimeoutError as bootup_error:
      raise errors.DeviceError(
          f"{self._device_name} failed to boot up within {bootup_timeout}s."
      ) from bootup_error

  @decorators.CapabilityLogDecorator(logger)
  def get_device_info(self) -> device_service_pb2.DeviceInfo:
    """Returns device static information."""
    payload_in_bytes = self._trigger_device_action(action="GetDeviceInfo")
    return device_service_pb2.DeviceInfo.FromString(payload_in_bytes)

  @decorators.CapabilityLogDecorator(logger)
  def get_device_state(self) -> device_service_pb2.DeviceState:
    """Returns device state information."""
    payload_in_bytes = self._trigger_device_action(action="GetDeviceState")
    return device_service_pb2.DeviceState.FromString(payload_in_bytes)

  @decorators.CapabilityLogDecorator(logger)
  def set_pairing_info(
      self, code: Optional[int] = None, discriminator: Optional[int] = None
    ) -> None:
    """Sets the pairing info of the device.

    Args:
      code: New pairing code to set.
      discriminator: New discriminator to set.
    """
    new_code = self.pairing_info.code if code is None else code
    new_discriminator = (
        self.pairing_info.discriminator if discriminator is None
        else discriminator)
    self._trigger_device_action(action="SetPairingInfo",
                                code=new_code,
                                discriminator=new_discriminator)

  @decorators.CapabilityLogDecorator(logger)
  def get_spake_info(self) -> device_service_pb2.SpakeInfo:
    """Returns the device spake information."""
    payload_in_bytes = self._trigger_device_action(action="GetSpakeInfo")
    return device_service_pb2.SpakeInfo.FromString(payload_in_bytes)

  @decorators.CapabilityLogDecorator(logger)
  def set_spake_info(
      self,
      verifier: Optional[bytes] = None,
      salt: Optional[bytes] = None,
      iteration_count: Optional[int] = None) -> None:
    """Sets the device spake information."""
    current_spake_info = self.get_spake_info()
    new_verifier = current_spake_info.verifier if verifier is None else verifier
    new_salt = current_spake_info.salt if salt is None else salt
    new_iteration_count = (
        current_spake_info.iteration_count if iteration_count is None
        else iteration_count)
    self._trigger_device_action(action="SetSpakeInfo",
                                verifier=new_verifier,
                                salt=new_salt,
                                iteration_count=new_iteration_count)

  @decorators.DynamicProperty
  def is_advertising(self) -> bool:
    """Returns if the device is advertising for commissioning."""
    payload_in_bytes = self._trigger_device_action(action="GetPairingState")
    payload = device_service_pb2.PairingState.FromString(payload_in_bytes)
    return payload.pairing_enabled

  @decorators.CapabilityLogDecorator(logger)
  def start_advertising(self) -> None:
    """Starts advertising for commissioning.

    Noop if device is already paired or it's already advertising.
    """
    if self.pairing_state or self.is_advertising:
      return
    self._trigger_device_action(action="SetPairingState",
                                pairing_enabled=True)

  @decorators.CapabilityLogDecorator(logger)
  def stop_advertising(self) -> None:
    """Stops advertising for commissioning.

    Noop if device is not advertising.
    """
    if not self.is_advertising:
      return
    self._trigger_device_action(action="SetPairingState",
                                pairing_enabled=False)

  @decorators.CapabilityLogDecorator(logger)
  def set_ota_metadata(self, tlv_metadata: bytes) -> None:
    """Sets OTA metadata for OTA provider.

    Note that the value does not persist across reboot. Also the device needs
    to be commissioned to the OTA provider (hub) to set the value.

    Args:
      tlv_metadata:
        TLV OTA metadata generated via gazoo_device.utility.tlv_utils.
    """
    self._trigger_device_action(action="SetOtaMetadataForProvider",
                                tlv=tlv_metadata)

  def _trigger_device_action(self, action: str, **kwargs: Any) -> bytes:
    """Triggers specific device action.

    Args:
      action: Device actions including reboot, factory-reset,
        OTA and setting/getting device state and pairing state.
      **kwargs: Arguments for the device action.

    Returns:
      The payload in bytes format.

    Raises:
      DeviceError: The ack status is not true.
    """
    device_rpc_kwargs = {"pw_rpc_timeout_s": self._rpc_timeout_s}
    device_rpc_kwargs.update(kwargs)
    return self.call(
        method_name=pigweed_rpc_transport.RPC_METHOD_NAME,
        method_args=("Device", action),
        method_kwargs=device_rpc_kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility module for interaction with ssh."""

import contextlib
import os
import subprocess
import tempfile
from typing import Generator

from gazoo_device import data_types
from gazoo_device import gdm_logger
from gazoo_device.utility import host_utils
from gazoo_device.utility import retry

logger = gdm_logger.get_logger()


@contextlib.contextmanager
def port_forward(username: str,
                 address: str,
                 key_info: data_types.KeyInfo,
                 remote_port: int,
                 remote_host: str = "localhost") -> Generator[str, None, None]:
  """Initializes an SSH port forward.

  Args:
    username: username which scp should use.
    address: IP address of the device.
    key_info: SSH key to use.
    remote_port: Remote port to forward to.
    remote_host: Remote host to forward to.

  Yields:
    The path of local unix socket which is forwarded to the remote.
  """
  with tempfile.TemporaryDirectory() as tempdir:
    local = f"{tempdir}/ssh_forward"
    remote = f"{remote_host}:{remote_port}"
    key_args = (["-i", host_utils.get_key_path(key_info)]
                if key_info is not None else [])
    ssh_command = [
        "ssh", "-N", *host_utils.DEFAULT_SSH_OPTIONS, *key_args,
        f"{username}@{address}", f"-L{local}:{remote}"
    ]
    logger.debug(f"Forking ssh command {ssh_command}")

    with subprocess.Popen(ssh_command) as process:
      try:
        retry.retry(lambda: os.path.exists(local), is_successful=lambda x: x)
        yield local
      finally:
        process.terminate()
        try:
          process.wait(timeout=1)
        except subprocess.TimeoutExpired:
          process.kill()
          raise



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Raspberry Pi Matter Controller device class."""

from typing import Optional

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.auxiliary_devices import raspberry_pi
from gazoo_device.base_classes import matter_endpoints_mixin
from gazoo_device.capabilities import matter_controller_chip_tool
from gazoo_device.capabilities import matter_endpoints_accessor_chip_tool

logger = gdm_logger.get_logger()

_LOGGING_CMD = ("tail", "-F", matter_controller_chip_tool.LOGGING_FILE_PATH)


class RaspberryPiMatterController(
    raspberry_pi.RaspberryPi,
    matter_endpoints_mixin.MatterEndpointAliasesMixin):
  """Base Class for RaspberryPiMatterController Devices."""
  _COMMUNICATION_KWARGS = {
      "log_cmd": _LOGGING_CMD,
      "key_info": config.KEYS["raspberrypi3_ssh_key"],
      "username": "pi"
  }
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SshQuery.IS_RASPBIAN_RPI: True,
      detect_criteria.SshQuery.IS_CHIP_TOOL_PRESENT: True,
  }
  DEVICE_TYPE = "rpi_matter_controller"
  _OWNER_EMAIL = "gdm-authors@google.com"

  @decorators.LogDecorator(logger)
  def factory_reset(self) -> None:
    """Factory resets the device."""
    self.matter_controller.factory_reset()

  @decorators.CapabilityDecorator(
      matter_controller_chip_tool.MatterControllerChipTool)
  def matter_controller(
      self) -> matter_controller_chip_tool.MatterControllerChipTool:
    """Matter controller capability to send chip-tool commands to the device."""
    return self.lazy_init(
        matter_controller_chip_tool.MatterControllerChipTool,
        device_name=self.name,
        regex_shell_fn=self.shell_with_regex,
        shell_fn=self.shell,
        send_file_to_device=self.file_transfer.send_file_to_device,
        get_property_fn=self.get_property,
        set_property_fn=self.set_property)

  @decorators.OptionalProperty
  def matter_node_id(self) -> Optional[int]:
    """Matter Node ID assigned to the currently commissioned end device."""
    return self.props["optional"].get("matter_node_id")

  @decorators.CapabilityDecorator(
      matter_endpoints_accessor_chip_tool.MatterEndpointsAccessorChipTool)
  def matter_endpoints(
      self
  ) -> matter_endpoints_accessor_chip_tool.MatterEndpointsAccessorChipTool:
    """Matter capability to access commissioned device's endpoint instances."""
    if self.matter_node_id is None:
      raise errors.DeviceError(
          "matter_endpoints requires a commissioned end device.")

    return self.lazy_init(
        matter_endpoints_accessor_chip_tool.MatterEndpointsAccessorChipTool,
        device_name=self.name,
        node_id_getter=lambda: self.matter_node_id,
        shell_fn=self.shell,
        shell_with_regex=self.shell_with_regex,
        matter_controller=self.matter_controller)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""M5stick dev board device class."""

from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import espressif_esp32_device
from gazoo_device.switchboard import switchboard

logger = gdm_logger.get_logger()


class M5Stick(espressif_esp32_device.EspressifESP32Device):
  """Device class for M5stick dev board.

  M5StickC is a mini M5Stack, powered by ESP32. It is a portable, easy-to-use,
  open source, IoT development board. https://docs.m5stack.com/en/core/m5stickc
  """
  COMMUNICATION_TYPE = "SerialComms"
  _COMMUNICATION_KWARGS = {}
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SerialQuery.ADDRESS: r"/dev/serial/by-id/*",
      detect_criteria.SerialQuery.PRODUCT_NAME: "m5stack",
  }
  DEVICE_TYPE = "m5stick"
  _OWNER_EMAIL = "gdm-authors@google.com"

  @decorators.CapabilityDecorator(switchboard.SwitchboardDefault)
  def switchboard(self):
    """Instance for communicating with the device."""
    switchboard_name = self._get_private_capability_name(
        switchboard.SwitchboardDefault)
    if not hasattr(self, switchboard_name):
      switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
      switchboard_kwargs.update({
          "communication_address": self.communication_address,
          "communication_type": self.COMMUNICATION_TYPE,
          "log_path": self.log_file_name,
          "device_name": self.name,
          "event_parser": None})
      setattr(self, switchboard_name,
              self.get_manager().create_switchboard(**switchboard_kwargs))

    return getattr(self, switchboard_name)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of the Matter Fan Control cluster capability.
"""
from typing import Union
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import fan_control_base
from gazoo_device.protos import attributes_service_pb2

FanControlCluster = matter_enums.FanControlCluster
ENUM8_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_ENUM8_ATTRIBUTE_TYPE
UINT8_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_INT8U_ATTRIBUTE_TYPE


class FanControlClusterPwRpc(fan_control_base.FanControlClusterBase):
  """Matter Fan Control cluster capability."""

  @decorators.DynamicProperty
  def fan_mode(self) -> matter_enums.FanMode:
    """The FanMode attribute."""
    return matter_enums.FanMode(self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_FAN_MODE,
        attribute_type=ENUM8_ATTRIBUTE_TYPE))

  @fan_mode.setter
  def fan_mode(self, value: matter_enums.FanMode) -> None:
    """Sets the FanMode attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_FAN_MODE,
        value=value,
        attribute_type=ENUM8_ATTRIBUTE_TYPE)

  @decorators.DynamicProperty
  def fan_mode_sequence(self) -> matter_enums.FanModeSequence:
    """The FanModeSequence attribute."""
    return matter_enums.FanModeSequence(self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_FAN_MODE_SEQUENCE,
        attribute_type=ENUM8_ATTRIBUTE_TYPE))

  @fan_mode_sequence.setter
  def fan_mode_sequence(self, value: matter_enums.FanModeSequence) -> None:
    """Sets the FanModeSequence attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_FAN_MODE_SEQUENCE,
        value=value,
        attribute_type=ENUM8_ATTRIBUTE_TYPE)

  @decorators.DynamicProperty
  def percent_setting(self) -> int:
    """The PercentSetting attribute."""
    return self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_PERCENT_SETTING)

  @percent_setting.setter
  def percent_setting(self, value: int) -> None:
    """Sets the PercentSetting attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_FAN_MODE_SEQUENCE,
        value=value)

  @decorators.DynamicProperty
  def percent_current(self) -> int:
    """The PercentCurrent attribute."""
    return self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_PERCENT_CURRENT)

  @percent_current.setter
  def percent_current(self, value: int) -> None:
    """Sets the PercentCurrent attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_PERCENT_CURRENT,
        value=value)

  @decorators.DynamicProperty
  def speed_max(self) -> int:
    """The SpeedMax attribute."""
    return self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_SPEED_MAX)

  @speed_max.setter
  def speed_max(self, value: int) -> None:
    """Sets the SpeedMax attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_SPEED_MAX,
        value=value)

  @decorators.DynamicProperty
  def speed_setting(self) -> int:
    """The SpeedSetting attribute."""
    return self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_SPEED_SETTING)

  @speed_setting.setter
  def speed_setting(self, value: int) -> None:
    """Sets the SpeedSetting attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_SPEED_SETTING,
        value=value)

  @decorators.DynamicProperty
  def speed_current(self) -> int:
    """The SpeedCurrent attribute."""
    return self._read_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_SPEED_CURRENT)

  @speed_current.setter
  def speed_current(self, value: int) -> None:
    """Sets the SpeedCurrent attribute."""
    self._write_attribute(
        attribute_id=FanControlCluster.ATTRIBUTE_SPEED_CURRENT,
        value=value)

  def _read_attribute(
      self, attribute_id: int, attribute_type: int = UINT8_ATTRIBUTE_TYPE
      ) -> int:
    """Reads the value from the attribute ID on the Fan Control cluster."""
    data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=FanControlCluster.ID,
        attribute_id=attribute_id,
        attribute_type=attribute_type)
    return data.data_uint8

  def _write_attribute(
      self,
      attribute_id: int,
      value: Union[int, matter_enums.FanMode, matter_enums.FanModeSequence],
      attribute_type: int = UINT8_ATTRIBUTE_TYPE):
    """Writes the value to the attribute ID on the Fan Control cluster."""
    self._write(endpoint_id=self._endpoint_id,
                cluster_id=FanControlCluster.ID,
                attribute_id=attribute_id,
                attribute_type=attribute_type,
                data_uint8=value)
    if self._read_attribute(attribute_id, attribute_type) != value:
      raise errors.DeviceError(
          f"Device {self._device_name} attribute {attribute_id} didn't change "
          f"to {value}.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter endpoint and cluster modules."""

from gazoo_device.capabilities.matter_clusters import basic_information_chip_tool
from gazoo_device.capabilities.matter_clusters import boolean_state_pw_rpc
from gazoo_device.capabilities.matter_clusters import color_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import door_lock_chip_tool
from gazoo_device.capabilities.matter_clusters import door_lock_pw_rpc
from gazoo_device.capabilities.matter_clusters import fan_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import flow_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import flow_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import illuminance_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import illuminance_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import level_control_chip_tool
from gazoo_device.capabilities.matter_clusters import level_control_pw_rpc
from gazoo_device.capabilities.matter_clusters import occupancy_sensing_chip_tool
from gazoo_device.capabilities.matter_clusters import occupancy_sensing_pw_rpc
from gazoo_device.capabilities.matter_clusters import on_off_chip_tool
from gazoo_device.capabilities.matter_clusters import on_off_pw_rpc
from gazoo_device.capabilities.matter_clusters import pressure_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import pressure_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import relative_humidity_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import relative_humidity_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import temperature_measurement_chip_tool
from gazoo_device.capabilities.matter_clusters import temperature_measurement_pw_rpc
from gazoo_device.capabilities.matter_clusters import thermostat_chip_tool
from gazoo_device.capabilities.matter_clusters import thermostat_pw_rpc
from gazoo_device.capabilities.matter_clusters import window_covering_pw_rpc
from gazoo_device.capabilities.matter_endpoints import color_temperature_light
from gazoo_device.capabilities.matter_endpoints import contact_sensor
from gazoo_device.capabilities.matter_endpoints import dimmable_light
from gazoo_device.capabilities.matter_endpoints import door_lock
from gazoo_device.capabilities.matter_endpoints import extended_color_light
from gazoo_device.capabilities.matter_endpoints import fan
from gazoo_device.capabilities.matter_endpoints import flow_sensor
from gazoo_device.capabilities.matter_endpoints import heating_cooling_unit
from gazoo_device.capabilities.matter_endpoints import humidity_sensor
from gazoo_device.capabilities.matter_endpoints import light_sensor
from gazoo_device.capabilities.matter_endpoints import occupancy_sensor
from gazoo_device.capabilities.matter_endpoints import on_off_light
from gazoo_device.capabilities.matter_endpoints import on_off_light_switch
from gazoo_device.capabilities.matter_endpoints import on_off_plugin_unit
from gazoo_device.capabilities.matter_endpoints import pressure_sensor
from gazoo_device.capabilities.matter_endpoints import root_node
from gazoo_device.capabilities.matter_endpoints import speaker
from gazoo_device.capabilities.matter_endpoints import temperature_sensor
from gazoo_device.capabilities.matter_endpoints import thermostat
from gazoo_device.capabilities.matter_endpoints import window_covering
import immutabledict


# PwRPC and CHIP tool implementation share the same endpoint interfaces.
SUPPORTED_ENDPOINTS = (
    color_temperature_light.ColorTemperatureLightEndpoint,
    contact_sensor.ContactSensorEndpoint,
    dimmable_light.DimmableLightEndpoint,
    door_lock.DoorLockEndpoint,
    extended_color_light.ExtendedColorLightEndpoint,
    fan.FanEndpoint,
    flow_sensor.FlowSensorEndpoint,
    heating_cooling_unit.HeatingCoolingUnitEndpoint,
    humidity_sensor.HumiditySensorEndpoint,
    light_sensor.LightSensorEndpoint,
    occupancy_sensor.OccupancySensorEndpoint,
    on_off_light.OnOffLightEndpoint,
    on_off_light_switch.OnOffLightSwitchEndpoint,
    on_off_plugin_unit.OnOffPluginUnitEndpoint,
    pressure_sensor.PressureSensorEndpoint,
    root_node.RootNodeEndpoint,
    speaker.SpeakerEndpoint,
    temperature_sensor.TemperatureSensorEndpoint,
    thermostat.ThermostatEndpoint,
    window_covering.WindowCoveringEndpoint)

SUPPORTED_CLUSTERS_PW_RPC = (
    color_control_pw_rpc.ColorControlClusterPwRpc,
    boolean_state_pw_rpc.BooleanStateClusterPwRpc,
    door_lock_pw_rpc.DoorLockClusterPwRpc,
    fan_control_pw_rpc.FanControlClusterPwRpc,
    flow_measurement_pw_rpc.FlowMeasurementClusterPwRpc,
    illuminance_measurement_pw_rpc.IlluminanceMeasurementClusterPwRpc,
    level_control_pw_rpc.LevelControlClusterPwRpc,
    occupancy_sensing_pw_rpc.OccupancySensingClusterPwRpc,
    on_off_pw_rpc.OnOffClusterPwRpc,
    pressure_measurement_pw_rpc.PressureMeasurementClusterPwRpc,
    (relative_humidity_measurement_pw_rpc.
     RelativeHumidityMeasurementClusterPwRpc),
    temperature_measurement_pw_rpc.TemperatureMeasurementClusterPwRpc,
    thermostat_pw_rpc.ThermostatClusterPwRpc,
    window_covering_pw_rpc.WindowCoveringClusterPwRpc)

MATTER_DEVICE_TYPE_ID_TO_CLASS = immutabledict.immutabledict({
    endpoint_class.DEVICE_TYPE_ID: endpoint_class
    for endpoint_class in SUPPORTED_ENDPOINTS
})

CLUSTER_ID_TO_CLASS_PW_RPC = immutabledict.immutabledict({
    cluster_class.CLUSTER_ID: cluster_class
    for cluster_class in SUPPORTED_CLUSTERS_PW_RPC
})

SUPPORTED_CLUSTERS_CHIP_TOOL = (
    basic_information_chip_tool.BasicInformationClusterChipTool,
    door_lock_chip_tool.DoorLockClusterChipTool,
    flow_measurement_chip_tool.FlowMeasurementClusterChipTool,
    illuminance_measurement_chip_tool.IlluminanceMeasurementClusterChipTool,
    level_control_chip_tool.LevelControlClusterChipTool,
    occupancy_sensing_chip_tool.OccupancySensingClusterChipTool,
    on_off_chip_tool.OnOffClusterChipTool,
    pressure_measurement_chip_tool.PressureMeasurementClusterChipTool,
    relative_humidity_measurement_chip_tool.
    RelativeHumidityMeasurementClusterChipTool,
    temperature_measurement_chip_tool.TemperatureMeasurementClusterChipTool,
    thermostat_chip_tool.ThermostatClusterChipTool,
)


CLUSTER_ID_TO_CLASS_CHIP_TOOL = immutabledict.immutabledict({
    cluster_class.CLUSTER_ID: cluster_class
    for cluster_class in SUPPORTED_CLUSTERS_CHIP_TOOL
})



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets information on all usb connected devices.

Returns a dictionary mapping a persistent address (adb or serial port) to usb
info for that address.

Note: usb_info_linux primarily relies on pyudev for full information.
It processes just adb devices and those that show up in /dev/serial/by-id
"""
import dataclasses
import re
import sys
from typing import Dict, List, Optional, Tuple

from gazoo_device.utility import usb_config
import immutabledict
import pyudev

PERSISTENT_SERIAL_PATH_FOLDER = '/dev/serial/by-id'
MOUNTED_DISK_FOLDER = '/dev/disk/by-id'
ADB_SERIAL_REGEX = r'\/dev\/disk\/by-id\/usb-Linux_File-CD_Gadget_([\da-z]+)-0:0'
ANDROID_DEVICE_PRODUCT_NAME = 'File-CD Gadget'

# Type aliases
_AddressStr = str
_AddressUsbInfoDict = Dict[_AddressStr, usb_config.UsbInfo]
_DiskInfo = Dict[str, str]
_DevPathStr = str

################################################################################
# configuration example.1:
#     Host -> Cambrionix_USB2 -+-> Pixel
#                              +-> Another serial device
#
#     Cambrionix_USB2 dev path:
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.4/1-14.4.1/1-14.4.1:1.0/ttyUSB0/tty/ttyUSB0"
#     Pixel dev path (no "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.2/1-14.2.1"
#     Another serial device dev path (with "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.2/1-14.2.2/1-14.2.2.1"
#
#     _USB2_HUB_IDENTIFIER_REGEX below finds the Cambrionix hub identifier.
#     (1) finds between "/" just before "ttyUSB0" which is "1-14.4.1:1.0" above.
#     (2) strip N.N:N.N at end which becomes
#         the Cambrionix hub identifier: 1-14.
#     (3) replace the hub_identifier inside _IDENTIFIER_REGEX_TEMPLATE by 1-14.
#
#     _IDENTIFIER_REGEX_TEMPLATE below finds the two digits following
#     the hub identifier in their dev path as key to look up
#     the USB2 Cambrionix port table.
#     For Pixel from "1-14.2.1" port key is "2.1" which is port number 15
#     in USB2 table.
#     For Another serial device from "1-14.2.2" port key is "2.2" which is
#     port number 14 in USB2 table.
#
# configuration example.2: Daisy chain
#     Host -> Cambrionix_USB3 -> Cambrionix_USB2 -+-> Pixel
#                                                 +-> Another serial device
#
#     Cambrionix_USB3 dev path:
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.5/1-14.5:1.1/tty/ttyACM0"
#     Cambrionix_USB2 dev path:
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.2/1-14.2.2/1-14.2.2.4/1-14.2.2.4.1/1-14.2.2.4.1:1.0"
#         "/ttyUSB0/tty/ttyUSB0"
#     Pixel dev path (no "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.2/1-14.2.2/1-14.2.2.2/1-14.2.2.2.1"
#     Another serial device dev path (with "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.2/1-14.2.2/1-14.2.2.2/1-14.2.2.2.2/1-14.2.2.2.2.1"
#
#     NOTICE: Cambrionix_USB2, Pixel and Another serial device dev path
#             get longer than non daisy chain.
#
#     _USB3_HUB_IDENTIFIER_REGEX below finds the USB3 Cambrionix hub identifier
#     which is 1-14.
#     The detail is described at _USB3_HUB_IDENTIFIER_REGEX comment.
#
#     _IDENTIFIER_REGEX_TEMPLATE below finds the two digits:2.2 following
#     the hub identifier:1-14 in Cambrionix_USB2 dev path as key to look up
#     the USB3 Cambrionix port table. From "/1-14.2.2/" port key is "2.2"
#     which is port number 7 at USB3 table. It means that port 0 of
#     Cambrionix USB2 is connected to port 7 of Cambrionix USB3.
#
#     _USB2_HUB_IDENTIFIER_REGEX below finds the USB2 Cambrionix hub identifier.
#     (1) finds between "/" just before "ttyUSB0" which is "1-14.2.2.4.1:1.0"
#         above.
#     (2) strip N.N:N.N at end which becomes
#         the USB2 Cambrionix hub identifier: 1-14.2.2.
#         Notice it is longer than configuration example.1.
#     (3) replace the hub_identifier inside _IDENTIFIER_REGEX_TEMPLATE
#         by 1-14.2.2.
#
#     _IDENTIFIER_REGEX_TEMPLATE below finds the two digits following
#     the hub identifier in their dev path as key to look up
#     the USB2 Cambrionix port table.
#     For Pixel from"1-14.2.2.2.1" port key is "2.1" which is port number 15
#     in USB2 table.
#     For Another serial device from "1-14.2.2.2.2" port key is "2.2"
#     which is port number 14 in USB2 table.
################################################################################
_USB2_HUB_IDENTIFIER_REGEX = r'''
##########################################################################
# The regexes below search for the longest "N-N.N.N.N.N.N" sequence,
# which corresponds to the USB hierarchy/path.
# RE extracts variable length hub identifier from devpath, then
# it replaces part of _IDENTIFIER_REGEX_TEMPLATE.
##########################################################################
  /                                # start by '/'

  # find hub identifier: For example,
  # '1-14' at configuration example.1
  # '1-14.2.2' at configuration example.2 for daisy chained device.
  # This sequence gets longer when daisy chain gets longer.
  (?P<hub_id>\d-\d+(?:\.\d+)*)

  # followed by '.N.N:N.N/tty' for USB2. USB3 is one digit shorter.
  \.\d+\.\d+:\d+\.\d+/tty
'''

################################################################################
# configuration example.3:
#     Host -> Cambrionix_USB3 -+-> Pixel
#                              +-> Another serial device
#
#     Cambrionix_USB3 dev path:
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.5/1-14.5:1.1/tty/ttyACM0"
#     Pixel dev path (no "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1/1-14/1-14.2/1-14.2.3"
#     Another serial device dev path (with "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1/1-14/1-14.2/1-14.2.4"
#         "/1-14.2.4.1"
#
#     _USB3_HUB_IDENTIFIER_REGEX below finds the Cambrionix hub identifier.
#     (1) finds between "/" just before "tty" which is "1-14.5:1.1" above.
#         Notice one digit shorter than Cambrionix USB2.
#     (2) strip N:N.N at end which becomes the Cambrionix hub identifier: 1-14.
#         Notice one digit shorter than Cambrionix USB2.
#     (3) replace the hub_identifier inside _IDENTIFIER_REGEX_TEMPLATE by 1-14.
#
#     _IDENTIFIER_REGEX_TEMPLATE below finds the two digits following
#     the hub identifier in their dev path as key to look up
#     the USB3 Cambrionix port table.
#     For Pixel from "1-14.2.3" port key is "2.3" which is port number 15
#     in USB3 table.
#     For Another serial device from "1-14.2.4" port key is "2.4"
#     which is port number 14 in USB3 table.
#
# configuration example.4: Daisy chain
#     Host -> Cambrionix_USB2 -> Cambrionix_USB3 -+-> Pixel
#                                                 +-> Another serial device
#
#     Cambrionix_USB2 dev path:
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.4/1-14.4.1/1-14.4.1:1.0/ttyUSB0/tty/ttyUSB0"
#     Cambrionix_USB3 dev path:
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.1/1-14.1.4/1-14.1.4.5/1-14.1.4.5:1.1/tty/ttyACM0"
#     Pixel dev path (no "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.1/1-14.1.4/1-14.1.4.2/1-14.1.4.2.3"
#     Another serial device dev path (with "DEVLINKS"):
#         "/devices/pci0000:00/0000:00:14.0/usb1"
#         "/1-14/1-14.1/1-14.1.4/1-14.1.4.2/1-14.1.4.2.4/1-14.1.4.2.4.1"
#
#     NOTICE: Cambrionix_USB3, Pixel and Another serial device dev path
#             get longer than configuration example.3.
#
#     _USB2_HUB_IDENTIFIER_REGEX below finds the USB2 Cambrionix hub identifier
#     which is 1-14.
#     The detail is described at _USB2_HUB_IDENTIFIER_REGEX comment.
#
#     _IDENTIFIER_REGEX_TEMPLATE below finds the two digits:2.2 following
#     the hub identifier:1-14 in Cambrionix_USB3 dev path as key to look up
#     the USB2 Cambrionix port table. From "/1-14.1.4/" port key is "1.4"
#     which is port number 7 at USB2 table. It means that port 0 of
#     Cambrionix USB3 is connected to port 7 of Cambrionix USB2.
#
#     _USB3_HUB_IDENTIFIER_REGEX below finds the USB3 Cambrionix hub identifier.
#     (1) finds between "/" just before "tty" which is "1-14.1.4.5:1.1"
#         above.
#     (2) strip N:N.N at end which becomes
#         the USB3 Cambrionix hub identifier: 1-14.1.4.
#         Notice it is longer than configuration example.3.
#     (3) replace the hub_identifier inside _IDENTIFIER_REGEX_TEMPLATE
#         by 1-14.1.4.
#
#     _IDENTIFIER_REGEX_TEMPLATE below finds the two digits following
#     the hub identifier in their dev path as key to look up
#     the USB3 Cambrionix port table.
#     For Pixel from "1-14.1.4.2.3" port key is "2.3" which is port number 15
#     in USB3 table.
#     For Another serial device from "1-14.1.4.2.4" port key is "2.4"
#     which is port number 14 in USB3 table.
################################################################################
_USB3_HUB_IDENTIFIER_REGEX = r'''
##########################################################################
# The regexes below search for the longest "N-N.N.N.N.N.N" sequence,
# which corresponds to the USB hierarchy/path.
# RE extracts variable length hub identifier from devpath, then
# it replaces part of _IDENTIFIER_REGEX_TEMPLATE.
##########################################################################
  /                                # start by '/'

  # find hub identifier: For example,
  # '1-14' at configuration example.3
  # '1-14.1.4' at configuration example.4 for daisy chained device.
  # This sequence gets longer when daisy chain gets longer.
  (?P<hub_id>\d-\d+(?:\.\d+)*)

  # followed by '.N:N.N/tty' for USB3. USB2 is one digit longer.
  \.\d+:\d+\.\d+/tty
'''

# The purpose of _IDENTIFIER_REGEX_TEMPLATE is find the port key:N.N from
# child dev path. Two of these numbers (each 'N' above) can be used to
# look up the physical port number from CAMBRIONIX_PORT_MAP.
# The format is N.N appearing after hub_identifier.
# See the comment at _USB2_HUB_IDENTIFIER_REGEX and _USB3_HUB_IDENTIFIER_REGEX
# for the detail.
_IDENTIFIER_REGEX_TEMPLATE = r'''
  /                          # start by '/'
  {hub_identifier}           # this part is replaced before RE starts.
  \.                         # followed by '.'
  (?P<port_id>\d+\.\d+)      # N.N is a key to search CAMBRIONIX_PORT_MAP
  (?:/|$)                    # end by '/' or end of string.
'''


@dataclasses.dataclass(frozen=True)
class _UsbPathRegex:
  parent_hub_re: str
  child_template_re: str

_CAMBRIONIX_REGEX = immutabledict.immutabledict({
    'PP8S':
        _UsbPathRegex(_USB2_HUB_IDENTIFIER_REGEX,
                      _IDENTIFIER_REGEX_TEMPLATE),
    'PP15S':
        _UsbPathRegex(_USB2_HUB_IDENTIFIER_REGEX,
                      _IDENTIFIER_REGEX_TEMPLATE),
    'PS15-USB3':
        _UsbPathRegex(_USB3_HUB_IDENTIFIER_REGEX,
                      _IDENTIFIER_REGEX_TEMPLATE),
    'SuperSync15':
        _UsbPathRegex(_USB3_HUB_IDENTIFIER_REGEX,
                      _IDENTIFIER_REGEX_TEMPLATE),
    'U16S':
        _UsbPathRegex(_USB2_HUB_IDENTIFIER_REGEX,
                      _IDENTIFIER_REGEX_TEMPLATE),
})

# These vendor's mounted drives show up as multiple usb connections.
# We use the one with the symlink.
REPEAT_VENDORS = [
    'FTDI',
    'cambrionix',
    'Linux',
    'Android',
]


def get_address_to_usb_info_dict() -> _AddressUsbInfoDict:
  """Returns a dict of address to the usb information."""
  udev_devices = get_pyudev_list_of_devices()
  relevant_devices = [
      device for device in udev_devices
      if device.properties.get('ID_BUS') == 'usb'
  ]
  address_to_usb_info_dict = {}
  location_dict = {}
  model_dict = {}
  address_device_list = []
  for device in relevant_devices:
    address, entry, dev_path, model = _process_udev_device(device)
    if address:  # remove devices that don't have good addresses
      address_device_list.append((address, device))
      address_to_usb_info_dict[address] = entry
      if dev_path:
        location_dict[address] = dev_path
      if model:
        model_dict[address] = model
  disk_info = _get_disk_info()
  _add_disk_info(address_to_usb_info_dict, disk_info)
  _add_cambrionix_ports(address_to_usb_info_dict, location_dict, model_dict,
                        address_device_list)
  return address_to_usb_info_dict


def get_pyudev_list_of_devices(
    subsystem: Optional[str] = None,
    devtype: Optional[str] = None
) -> List[pyudev.Device]:
  """Wrapper around pyudev.list_devices.

  Args:
    subsystem: maps to pyudev.Context().list_devices arg.
    devtype: maps to pyudev.Context().list_devices arg DEVTYPE.

  Returns:
    a list of pyudev devices.

  Raises:
    RuntimeError: if platform is not linux.
  """
  args = {}
  if subsystem:
    args['subsystem'] = subsystem
  if devtype:
    args['DEVTYPE'] = devtype
  if sys.platform == 'darwin':
    raise RuntimeError('This method is only supported on linux.')
  context = pyudev.Context()
  return list(context.list_devices(**args))


def _add_cambrionix_ports(
    address_to_usb_info_dict: _AddressUsbInfoDict,
    location_dict: Dict[_AddressStr, _DevPathStr],
    model_dict: Dict[_AddressStr, str],
    address_device_list: List[tuple[_AddressStr, pyudev.Device]]
) -> None:
  """Add cambrionix port number and parent address to each cambrionix child.

  (1) add children address of connected devices to cambrionix hub device.
  (2) add cambrionix port number and parent address to each cambrionix child.

  Args:
    address_to_usb_info_dict: Dictionary to look up the entry by address.
    location_dict: Dictionary to look up the dev_path by address.
    model_dict: Dictionary to look up the model by address.
    address_device_list: List of address and udev_device used to
        fill out child_addresses of Cambrionix hub device.
  """
  for address, udev_device in address_device_list:
    entry = address_to_usb_info_dict[address]
    if entry.product_name in usb_config.CAMBRIONIX_NAMES:
      entry.child_addresses = _get_child_addresses(udev_device,
                                                   address_to_usb_info_dict)
  for hub_address, model in model_dict.items():
    for child_address in address_to_usb_info_dict[hub_address].child_addresses:
      if child_address in address_to_usb_info_dict:
        child_entry = address_to_usb_info_dict[child_address]
        child_entry.usb_hub_address = hub_address
        child_entry.usb_hub_port = _get_cambrionix_port_number(
            hub_address, model, child_address, location_dict)


def _add_disk_info(
    address_to_usb_info_dict: _AddressUsbInfoDict,
    disk_info: _DiskInfo
) -> None:
  """Add disk path to the entries that have them.

  Args:
      address_to_usb_info_dict: Dictionary to look up the entry by address.
      disk_info: disk info
  """
  for entry in address_to_usb_info_dict.values():
    serial = entry.serial_number
    if serial in disk_info:
      entry.disk = disk_info[serial]


def _get_address(udev_device: pyudev.Device) -> str:
  """Gets the serial port symlink under '/dev/serial/by-id', mounted drives, or adb serial.

  Args:
      udev_device: udev device.

  Returns:
      the serial port symlink under '/dev/serial/by-id', mounted drives,
          or adb serial.
          if symlink is not found and is not known path, return ''.
  """
  devlinks_str = udev_device.properties.get('DEVLINKS', '')
  paths = devlinks_str.split()
  if not devlinks_str:  # Get mounted drives without corresponding symlinks
    if udev_device.properties.get('ID_VENDOR') not in REPEAT_VENDORS:
      return udev_device.properties.get('DEVNAME')

  for path in paths:
    if PERSISTENT_SERIAL_PATH_FOLDER in path:
      return path
    if MOUNTED_DISK_FOLDER in path:
      match = re.match(ADB_SERIAL_REGEX, path)
      if match:
        return match.group(1)
      return path

  return ''


def _get_cambrionix_model(udev_device: pyudev.Device) -> str:
  """Gets the model of the cambrionix.

  Args:
      udev_device: udev device.

  Returns:
      the model of the cambrionix
  """
  parent = udev_device.parent.parent.parent.parent
  if udev_device.properties.get('ID_MODEL') in usb_config.CAMBRIONIX_PORT_MAP:
    return udev_device.properties.get('ID_MODEL')
  else:
    return parent.properties.get('ID_MODEL')


def _get_cambrionix_port_number(
    hub_address: _AddressStr,
    hub_model: str,
    child_address: _AddressStr,
    location_dict: Dict[_AddressStr, _DevPathStr]
) -> Optional[int]:
  """Get port number from dev path.

  Args:
    hub_address: Address for the parent Cambrionix hub.
    hub_model: Model for the parent Cambrionix hub.
    child_address: Address for a child device connected to the hub.
    location_dict: Dictionary to look up the dev_path by address.

  Returns:
    Port number or None if not found.

  Note:
    The port number is determined by first finding the Cambionix (or parent)
    identifier. The parent identifier can vary in length, but is found using
    the hub dev path by looking between the '/' and the last one or two digits
    (depending on hub model) before the ':' ("/<parent_identifier>n.n:n.n").
    For children, the two digits following the parent identifier in their
    dev path are used to look up the Cambrionix port.

    Example:
      hub dev path:
        "/devices/pci0000:00/0000:00:14.0/usb1/1-4/1-4.1/1-4.1.2/1-4.1.2.4/"
        "1-4.1.2.4.1/1-4.1.2.4.1:1.0/ttyUSB4/tty/ttyUSB4"
      From the hub dev path "1-4.1.2.4.1:1.0" is used to find that the parent
      identifier is "1-4.1.2".
      child dev path:
        "/devices/pci0000:00/0000:00:14.0/usb1/1-4/1-4.1/1-4.1.2/1-4.1.2.3/"
        "1-4.1.2.3.4/1-4.1.2.3.4:1.2/ttyUSB2/tty/ttyUSB2"
          "1-4.1.2.3.4:1.2" is the relevant number.
          "1-4.1.2" is the parent identifier.
          "3.4" is the child identifier used to look up the port number.
          "1.2" indicates the FTDI interface.
      child dev path (hub device):
        "/devices/pci0000:00/0000:00:14.0/usb1/1-4/1-4.1/1-4.1.2/1-4.1.2.4/"
        "1-4.1.2.4.2/1-4.1.2.4.2.1/1-4.1.2.4.2.1:1.0/tty/ttyACM1"
          "1-4.1.2.4.2.1:1.0" is the relevant number.
          "1-4.1.2" is the parent identifier.
          "4.2" is the child identifier used to look up the port number.
          "1:" is the hub device identifier
          "1.0" indicates the FTDI interface.
  """
############################################################################
#
# Port finding flow:
#
# Model
#   |
#   |
#   +---> _CAMBRIONIX_REGEX dict
#   |        |
#   |        |                              search hub identifier
#   |        +---> parent_hub_re -----------------------+
#   |        |                                          |
#   |        |                                          V
#   |        |                       hub dev path "...<hub>............"
#   |        |                                          |
#   |        |                                          V
#   |        |          +-------------------------------+
#   |        |          |
#   |        |          |
#   |        |          | insert into child_template_re
#   |        |          |
#   |        |          |
#   |        V          V                              search child identifier
#   |        +---> child_template_re ------------------------------+
#   |                                                              |
#   |                                                              V
#   |                                child dev path "..<hub>...<child>......"
#   |                                                              |
#   V                                                              V
#   +--------+-----------------------------------------------------+
#            |
#            V
#     CAMBRIONIX_PORT_MAP
#            |
#            V
#        port number
#
############################################################################
  hub_id_regex = _CAMBRIONIX_REGEX[hub_model].parent_hub_re
  hub_match = re.search(hub_id_regex, location_dict[hub_address], re.VERBOSE)
  if hub_match:
    parent_hub_identifier = hub_match.group('hub_id').replace('.', r'\.')
    template_re = _CAMBRIONIX_REGEX[hub_model].child_template_re
    identifier_regex = template_re.format(hub_identifier=parent_hub_identifier)
    port_match = re.search(identifier_regex, location_dict[child_address],
                           re.VERBOSE)
    if port_match:
      port_mapping = usb_config.CAMBRIONIX_PORT_MAP[hub_model]
      return port_mapping.get(port_match.group('port_id'))  # port number
  return None


def _get_child_addresses(
    udev_device: pyudev.Device,
    address_to_usb_info_dict: _AddressUsbInfoDict
) -> List[_AddressStr]:
  """Get the addresses for the relevant children of the udev device.

  Args:
      udev_device: udev device of Cambrionix hub
      address_to_usb_info_dict: dictionary of address to UsbInfo instance
          for each usb device attached.

  Returns:
      a list of the addresses for the relevant children of
          the udev device.
  """
  all_children_devices = _get_cambrionix_udev_children(udev_device)
  relevant_children_devices = [
      child_device for child_device in all_children_devices
      if child_device != udev_device and
      _get_address(child_device) in address_to_usb_info_dict
  ]
  grand_children_devices = []
  for child_device in relevant_children_devices:
    # Remove children of other cambrionixes
    if _get_product_name(child_device) in usb_config.CAMBRIONIX_NAMES:
      grand_children_devices += _get_cambrionix_udev_children(child_device)
      grand_children_devices.remove(child_device)
  return [
      _get_address(child_device)
      for child_device in relevant_children_devices
      if child_device not in grand_children_devices
  ]


def _get_disk_info() -> Dict[str, str]:
  """Returns a dictionary of disk paths to serial numbers."""
  devices = get_pyudev_list_of_devices(subsystem='block')
  mounted_info = {
      device.properties.get('ID_SERIAL_SHORT', ''):
          device.properties.get('DEVNAME')
      for device in devices
  }
  return mounted_info


def _get_product_name(udev_device: pyudev.Device) -> str:
  """Returns the product name in a system agnostic way.

  Args:
      udev_device: udev device.
  """
  model = udev_device.properties.get('ID_MODEL', '').replace('_', ' ')
  if model == ANDROID_DEVICE_PRODUCT_NAME:  # Actually an Android device.
    model = 'Android'
  return model


def _process_udev_device(
    udev_device: pyudev.Device
) -> Tuple[_AddressStr, usb_config.UsbInfo, _DevPathStr, str]:
  """Get (address, UsbInfo, dev_path, model) from udev_device.

  Args:
      udev_device: udev device.

  Returns:
      tuple: (address, UsbInfo, dev_path, model)
      If address is invalid, return ('', UsbInfo, '', '') and
      the caller should ignore the returned tupple.
  """
  model = ''
  dev_path = udev_device.properties.get('DEVPATH', '')

  entry = usb_config.UsbInfo()
  address = _get_address(udev_device)
  if not address:
    return '', entry, '', ''
  entry.address = address
  entry.product_id = udev_device.properties.get('ID_MODEL_ID', '')
  entry.vendor_id = udev_device.properties.get('ID_VENDOR_ID', '')
  entry.ftdi_interface = int(
      udev_device.properties.get('ID_USB_INTERFACE_NUM', 0))
  entry.manufacturer = udev_device.properties.get('ID_VENDOR', '')
  entry.serial_number = udev_device.properties.get('ID_SERIAL_SHORT', '')
  entry.product_name = _get_product_name(udev_device)
  if entry.product_name in usb_config.CAMBRIONIX_NAMES:
    model = _get_cambrionix_model(udev_device)
  return address, entry, dev_path, model


def _get_cambrionix_udev_children(
    udev_device: pyudev.Device
) -> List[pyudev.Device]:
  """Gets all the udev devices representing a cambrionix's children.

  Args:
      udev_device: Cambrionix hub device.

  Returns:
      list of all the udev devices representing a cambrionix's children.

  Raises:
      NotImplementedError: if product name is not known.
  """
  product_name = _get_product_name(udev_device)
  if product_name in usb_config.CAMBRIONIX_NAMES_USB2:
    return list(udev_device.parent.parent.parent.parent.parent.children)
  if product_name in usb_config.CAMBRIONIX_NAMES_USB3:
    return list(udev_device.parent.parent.parent.parent.children)
  raise NotImplementedError(
      f'Unknown product name: {product_name} is found.')



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for the Matter Measurement cluster capability."""
import abc
import enum

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device.capabilities.matter_clusters.interfaces import cluster_base
from gazoo_device.protos import attributes_service_pb2

# According to the spec, MeasuredValue, MinMeasuredValue and MaxMeasuredValue
# attributes can only be int16 or uint16 type. The type logic is handled in
# the Ember API below.
_INT16SIGNED = attributes_service_pb2.AttributeType.ZCL_INT16S_ATTRIBUTE_TYPE


class MeasurementClusterBase(cluster_base.ClusterBase, metaclass=abc.ABCMeta):
  """Matter Pressure Measurement cluster capability."""

  CLUSTER_ID = None
  MATTER_CLUSTER = None
  ATTRIBUTE_TYPE = None

  @decorators.DynamicProperty
  def measured_value(self) -> int:
    """The MeasuredValue attribute.

    Returns:
      The MeasuredValue attribute.
    """
    return self._read_value(
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_MEASURED_VALUE)

  @measured_value.setter
  def measured_value(self, value: int) -> None:
    """Updates the MeasuredValue attribute with new value."""
    self._write_value(
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_MEASURED_VALUE, value=value)

  @decorators.DynamicProperty
  def min_measured_value(self) -> int:
    """The MinMeasuredValue attribute.

    The MinMeasuredValue attribute indicates the minimum value of MeasuredValue
    that is capable of being measured.

    Returns:
      The MinMeasuredValue attribute.
    """
    return self._read_value(
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_MIN_MEASURED_VALUE)

  @min_measured_value.setter
  def min_measured_value(self, value: int) -> None:
    """Updates the MinMeasuredValue attribute with new value."""
    self._write_value(
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_MIN_MEASURED_VALUE,
        value=value)

  @decorators.DynamicProperty
  def max_measured_value(self) -> int:
    """The MaxMeasuredValue attribute.

    The MaxMeasuredValue attribute indicates the maximum value of MeasuredValue
    that is capable of being measured.

    Returns:
      The MaxMeasuredValue attribute.
    """
    return self._read_value(
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_MAX_MEASURED_VALUE)

  @max_measured_value.setter
  def max_measured_value(self, value: int) -> None:
    """Updates the MaxMeasuredValue attribute with new value."""
    self._write_value(
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_MAX_MEASURED_VALUE,
        value=value)

  def _read_value(self, attribute_id: enum.IntEnum) -> int:
    """Reads the value from the given attribute ID.

    Args:
      attribute_id: Attribute ID on PressureMeasurementCluster.

    Returns:
      The value read from the attribute.
    """
    value_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=self.MATTER_CLUSTER.ID,
        attribute_id=attribute_id,
        attribute_type=self.ATTRIBUTE_TYPE)
    if self.ATTRIBUTE_TYPE == _INT16SIGNED:
      return value_data.data_int16
    else:
      return value_data.data_uint16

  def _write_value(self, attribute_id: enum.IntEnum, value: int) -> None:
    """Writes the value to the given attribute ID.

    Args:
      attribute_id: Attribute ID on PressureMeasurementCluster.
      value: Value to write to the attribute.

    Raises:
      DeviceError when the attribute value doesn't change.
    """
    if self.ATTRIBUTE_TYPE == _INT16SIGNED:
      self._write(
          endpoint_id=self._endpoint_id,
          cluster_id=self.MATTER_CLUSTER.ID,
          attribute_id=attribute_id,
          attribute_type=self.ATTRIBUTE_TYPE,
          data_int16=value)
    else:
      self._write(
          endpoint_id=self._endpoint_id,
          cluster_id=self.MATTER_CLUSTER.ID,
          attribute_id=attribute_id,
          attribute_type=self.ATTRIBUTE_TYPE,
          data_uint16=value)
    if self._read_value(attribute_id) != value:
      raise errors.DeviceError(
          f"Device {self._device_name} Attribute {attribute_id} didn't change"
          f" to {value}")


class MeasurementClusterChipToolBase(MeasurementClusterBase):
  """Matter Measurement cluster capability for chip-tool."""

  CLUSTER_NAME = None

  def _map_attribute_id_to_name(self, attribute_id: enum.IntEnum) -> str:
    """Converts attribute enum into chip-tool compatible attribute name.

    Args:
      attribute_id: Enum of the attribute. The enum name should start with
        ATTRIBUTE_. See list of supported attributes in matter_enums.

    Returns:
      Attribute name in kebab case.
      (e.g. ATTRIBUTE_MEASURED_VALUE -> measured-value)
    """
    attribute = attribute_id.name
    if not attribute.startswith("ATTRIBUTE_"):
      raise ValueError(
          f"Invalid attribute ID {attribute_id} for cluster {self.CLUSTER_ID}")
    return attribute.replace("ATTRIBUTE_", "").replace("_", "-").lower()

  def _read_value(self, attribute_id: enum.IntEnum) -> int:
    """Reads the value from the given attribute ID.

    Args:
      attribute_id: Attribute ID on a MeasurementCluster.

    Returns:
      The value read from the attribute.
    """
    return self._read(self._endpoint_id, self.CLUSTER_NAME,
                      self._map_attribute_id_to_name(attribute_id))

  def _write_value(self, attribute_id: enum.IntEnum, value: int) -> None:
    """Writes the value to the given attribute ID.

    Args:
      attribute_id: Attribute ID on a MeasurementCluster.
      value: Value to write to the attribute.

    Raises:
      DeviceError when the attribute value doesn't change.
    """
    self._write(self._endpoint_id, self.CLUSTER_NAME,
                self._map_attribute_id_to_name(attribute_id), [value])
    if self._read_value(attribute_id) != value:
      raise errors.DeviceError(
          f"Device {self._device_name} Attribute {attribute_id} didn't change"
          f" to {value}")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Contact Sensor endpoint.

This endpoint module corresponds to the
"Contact Sensor" Matter device type which defines required and optional
clusters in the Matter spec 1.0.

The required clusters for this endpoint: Boolean State
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import boolean_state_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import contact_sensor_base


class ContactSensorEndpoint(contact_sensor_base.ContactSensorBase):
  """Matter Contact Sensor endpoint."""

  @decorators.CapabilityDecorator(boolean_state_pw_rpc.BooleanStateClusterPwRpc)
  def boolean_state(self) -> boolean_state_pw_rpc.BooleanStateClusterPwRpc:
    """Matter Boolean State cluster instance."""
    return self.cluster_lazy_init(matter_enums.BooleanStateCluster.ID)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Window Covering endpoint.

This endpoint module corresponds to the
"Window Covering" Matter device type which defines required and optional
clusters in the Matter spec 1.0.

A Window Covering endpoint houses an instance of Window Covering Matter
device type with the following defined clusters on this endpoint:

The required clusters for this endpoint: Window Covering
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import window_covering_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import window_covering_base


class WindowCoveringEndpoint(window_covering_base.WindowCoveringBase):
  """Matter Window Covering endpoint."""

  @decorators.CapabilityDecorator(
      window_covering_pw_rpc.WindowCoveringClusterPwRpc)
  def window_covering(
      self
  ) -> window_covering_pw_rpc.WindowCoveringClusterPwRpc:
    """Matter Window Covering cluster instance."""
    return self.cluster_lazy_init(matter_enums.WindowCoveringCluster.ID)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Nrfjprog binary implementation for flashing NRF devices."""
import os
import tempfile
from typing import Callable, List, Optional, Tuple

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import flash_build_base
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.utility import host_utils
from gazoo_device.utility import subprocess_utils


logger = gdm_logger.get_logger()

# NRF command line tools
# https://www.nordicsemi.com/Products/Development-tools/nrf-command-line-tools
_NRFJPROG = "nrfjprog"
_FLASH_TIMEOUT_S = 180
_JLINK_EXE = "JLinkExe"


class FlashBuildNrfjprog(flash_build_base.FlashBuildBase):
  """Nrfjprog binary implementation for flashing NRF devices."""

  def __init__(
      self,
      device_name: str,
      serial_number: str,
      reset_endpoints_fn: Optional[Callable[[], None]] = None,
      switchboard: Optional[switchboard_base.SwitchboardBase] = None,
      wait_for_bootup_complete_fn: Optional[Callable[[int], None]] = None,
      power_cycle_fn: Optional[Callable[[], None]] = None):
    """Initializes an instance of the FlashBuildNrfjprog capability.

    Args:
      device_name: Device name used for logging.
      serial_number: Device serial number.
      reset_endpoints_fn: Method to reset matter_endpoint capability.
        This method will be called after flashing is completed.
      switchboard: A Switchboard capability instance if the device supports it.
      wait_for_bootup_complete_fn: wait_for_bootup_complete method for verifying
        device responsiveness after flashing.
      power_cycle_fn: Method to power cycle the device.
    """
    super().__init__(device_name=device_name)
    self._serial_number = serial_number
    self._reset_endpoints_fn = reset_endpoints_fn
    self._switchboard = switchboard
    self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
    self._power_cycle_fn = power_cycle_fn
    self._msd_disabled = False

  @decorators.CapabilityLogDecorator(logger, level=None)
  def health_check(self) -> None:
    """Checks whether the 'nrfjprog' binary is installed.

    Raises:
      DependencyUnavailableError: 'nrfjprog' is not installed.
    """
    if self._healthy is None:
      self._healthy = host_utils.has_command(_NRFJPROG)
    if not self._healthy:
      raise errors.DependencyUnavailableError(
          f"{_NRFJPROG} is not installed. Install the binary on the host from "
          "https://www.nordicsemi.com/Products/Development-tools/"
          "nrf-command-line-tools/download")
    self._disable_msd()

  def _disable_msd(self) -> None:
    """Disables the Mass Storage Device on host.

    Due to a known issue in Segger J-Link firmware, data corruption or drops
    will occur if using serial ports with packets larger than 64 bytes.
    Multiple RPC failures are also observed with this issue: b/255635727,
    b/261901521 and b/261846864.

    Disabling the MSD setting on host can fix this issue suggested by Segger:
    https://docs.zephyrproject.org/3.0.0/guides/tools/nordic_segger.html#disabling-the-mass-storage-device-functionality
    """
    if self._msd_disabled:
      return

    if not host_utils.has_command(_JLINK_EXE):
      raise errors.DependencyUnavailableError(
          f"{_JLINK_EXE} is not installed. Install the binary on the host from "
          "https://www.nordicsemi.com/Products/Development-tools/"
          "nrf-command-line-tools/download")

    # Create a jlink script for disabling MSD.
    with tempfile.TemporaryDirectory() as temp_directory:
      jlink_script = os.path.join(temp_directory, "msd.jlink")
      with open(jlink_script, "w") as f:
        f.write("MSDDisable\nExit\n")
      serial_number = self._serial_number.lstrip("0")
      return_code, output = subprocess_utils.run_and_stream_output(
          [_JLINK_EXE, "-CommandFile", jlink_script, "-USB", serial_number],
          timeout=5)
      if return_code:
        raise errors.DeviceError(
            f"{self._device_name} failed to disable the MSD setting. "
            f"Script return code: {return_code}, output: {output}")

    if self._power_cycle_fn is None:
      raise errors.DeviceError(
          f"{self._device_name} no power cycle method provided. Please unplug "
          "and replug the USB connection of the board manually.")

    # Try power cycling the device if it's connected to cambrionix.
    try:
      self._power_cycle_fn()
    except errors.CapabilityNotReadyError as err:
      raise errors.DeviceError(
          f"{self._device_name} cannot be powered cycle: {err}. Please unplug "
          "and replug the USB connection of the board manually.")

    self._msd_disabled = True

  @decorators.CapabilityLogDecorator(logger)
  def flash_device(self,
                   list_of_files: List[str],
                   expected_version: Optional[str] = None,
                   expected_build_type: Optional[str] = None,
                   verify_flash: bool = True,
                   method: Optional[str] = None) -> None:
    """Flashes the firmware image (.hex file) on the device.

    Args:
      list_of_files: Image files on local host, currently supports flashing
        only one hex file at a time.
      expected_version: Not used.
      expected_build_type: Not used.
      verify_flash: Check if we should verify build after flashing.
      method: Not used.

    Raises:
      ValueError: If invalid arguments are provided.
      DeviceError: If flashing fails.
      DependencyUnavailableError: If J-Link SDK is not installed.
    """
    del expected_version, expected_build_type, method  # Unused.

    self.health_check()

    # Check if only one valid hex file is given.
    if len(list_of_files) != 1:
      raise ValueError("Only one hex file can be flashed via JLink.")
    image_path = list_of_files[0]
    if not image_path.endswith(".hex"):
      raise ValueError("Only hex type file can be flashed.")
    if not os.path.exists(image_path):
      raise ValueError(f"Firmware image {image_path} does not exist.")

    if self._switchboard is not None:
      self._switchboard.close_all_transports()
    try:
      # Always recover the device before flashing to avoid flakiness
      self.recover_device()
      return_code, output = self._nrfjprog_flash(image_path)
    finally:
      if self._switchboard is not None:
        self._switchboard.open_all_transports()
      if self._reset_endpoints_fn is not None:
        self._reset_endpoints_fn()
      if verify_flash and self._wait_for_bootup_complete_fn is not None:
        self._wait_for_bootup_complete_fn()

    if return_code:
      raise errors.DeviceError(
          f"{self._device_name} flash command with binary flasher failed. "
          f"Return code: {return_code}. Output: {output!r}")

  def _nrfjprog_flash(self, image_path: str) -> Tuple[int, str]:
    """Flashes the device via nrfjprog binary if it is present."""
    snr = self._serial_number.lstrip("0")
    flash_command = (
        f"{_NRFJPROG} -f nrf52 --program {image_path} --chiperase -s {snr} "
        "--reset --verify --log nrfjprog_flashing_log.log")
    return subprocess_utils.run_and_stream_output(
        flash_command.split(), timeout=_FLASH_TIMEOUT_S)

  @decorators.CapabilityLogDecorator(logger)
  def recover_device(self) -> None:
    """Recovers the device by erasing all user available non-volatile memory."""
    self.health_check()

    serial_number = self._serial_number.lstrip("0")
    return_code, output = subprocess_utils.run_and_stream_output(
        [_NRFJPROG, "--family", "NRF52", "--recover", "--snr", serial_number],
        timeout=_FLASH_TIMEOUT_S)
    if return_code:
      raise errors.DeviceError(
          f"Failed to recover the device {serial_number}: {output}.")

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(self, build_file: str):
    """Upgrade the device based on the provided build arguments.

    Args:
      build_file: Local path to the file.
    """
    self.flash_device(list_of_files=[build_file])

  @decorators.CapabilityLogDecorator(logger)
  def extract_build_info(self, *args, **kwargs):
    """Converts the provided build arguments into info about the build."""
    raise NotImplementedError(
        "extract_build_info is not available in flash_build_nrfjprog for now.")
